/*-
 * Copyright (c) 2018 UPLEX Nils Goroll Systemoptimierung
 * All rights reserved
 *
 * Author: Geoffrey Simmons <geoffrey.simmons@uplex.de>
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

/*
 * Inspired in part by Varnish hash_critbit.c and tarsnap's Patricia
 * implementation.
 */

#include <stdlib.h>
#include <string.h>
#include <stdint.h>

#include "vdef.h"
#include "vas.h"
#include "miniobj.h"

#include "patricia.h"

/*
 * A trie is comprised by a tree of pt_y nodes and an array of strings,
 * both of which are owned by a VMOD object. Both are passed in to the
 * PT_* functions.
 *
 * idx is an index to the strings array, and the region of the string
 * represented at a node is strings[idx][off] to strings[idx][off+len],
 * inclusive.
 *
 * Terminating nulls form a part of the target string. So "foo\0" and
 * "foobar\0" differ at the fourth byte, and the critical bit at that byte
 * is the most significant 1-bit in 'b'.
 *
 * bitmask is set at the critical bit in the differing byte, and
 * determines whether searches continue into leaf[0] or leaf[1].
 */

struct pt_y {
	unsigned	magic;
#define PT_Y_MAGIC	0xfa564d14
	struct pt_y	*leaf[2];
	unsigned	idx;
	unsigned short	off;
	unsigned short	len;
	unsigned char	bitmask;
};

static unsigned char pt_bittbl[256] = { 0 };

void
PT_Init(void)
{
	unsigned char x;
	unsigned y;

	AZ(pt_bittbl[0x34 ^ 0x34]);

	y = 0;
	for (x = 0; x < 8; x++)
		for (; y < (1U << x); y++)
			pt_bittbl[y] = 8 - x;

	/* Quick asserts for sanity check */
	assert(pt_bittbl[0x34 ^ 0x34] == 8);
	AZ(pt_bittbl[0xaa ^ 0x55]);
	assert(pt_bittbl[0x01 ^ 0x22] == 2);
	assert(pt_bittbl[0x10 ^ 0x0b] == 3);
}

int
PT_Inited(void)
{
	return (pt_bittbl[0x34 ^ 0x34] != 0);
}

static struct pt_y *
y_alloc(unsigned idx, unsigned short off, size_t len)
{
	struct pt_y *y;

	if (len > USHRT_MAX) {
		errno = ERANGE;
		return (NULL);
	}

	errno = 0;
	ALLOC_OBJ(y, PT_Y_MAGIC);
	if (y == NULL)
		return (NULL);

	y->idx = idx;
	y->off = off;
	y->len = (unsigned short) len;
	AZ(y->leaf[0]);
	AZ(y->leaf[1]);
	AZ(y->bitmask);
	return (y);
}

static inline struct pt_y *
y_leaf_alloc(unsigned idx, unsigned char *c, unsigned char *b)
{
	return y_alloc(idx, (unsigned short)(uintptr_t)(c - b),
		       strlen((char *)c));
}

static struct pt_y *
y_dup(struct pt_y *y0, unsigned short len)
{
	struct pt_y *y;

	assert(len < y0->len);

	y = y_alloc(y0->idx, y0->off + len, y0->len - len);
	if (y == NULL)
		return (NULL);

	y->leaf[0] = y0->leaf[0];
	y->leaf[1] = y0->leaf[1];
	y->bitmask = y0->bitmask;
	return (y);
}

int
PT_Insert(struct pt_y * * restrict root, unsigned idx,
	  char * const restrict * const restrict strings)
{
	struct pt_y *y;
	unsigned char *c, *b;

	AN(root);
	CHECK_OBJ_ORNULL(*root, PT_Y_MAGIC);
	AN(strings);
	AN(strings[idx]);

	if (*root == NULL) {
		*root = y_alloc(idx, 0, strlen(strings[idx]));
		if (*root == NULL)
			return (-1);
		return (0);
	}

	y = *root;
	b = (unsigned char *)strings[idx];
	c = b;
	errno = 0;

	for (;;) {
		unsigned short i;
		unsigned char *s;
		unsigned char bit;
		struct pt_y *y_old, *y_new;

		CHECK_OBJ_NOTNULL(y, PT_Y_MAGIC);

		s = (unsigned char *)(strings[y->idx] + y->off);
		for (i = 0; *c != '\0' && i < y->len && s[i] == *c; i++)
			c++;

		if (s[i] == '\0' && *c == '\0') {
			/*
			 * The string to be inserted is already in the
			 * trie.
			 */
			assert(i == y->len);
			errno = EINVAL;
			return (-1);
		}

		if (i == y->len) {
			/*
			 * The string to be inserted has a prefix that is
			 * already in the trie.
			 */
			bit = (y->bitmask & *c) != 0;
			assert(bit < 2);
			if (y->leaf[bit] != NULL) {
				y = y->leaf[bit];
				continue;
			}
			y_new = y_leaf_alloc(idx, c, b);
			if (y_new == NULL)
				return (-1);
			y->leaf[bit] = y_new;
			return (0);
		}

		/* Split the current node. */

		y_new = y_leaf_alloc(idx, c, b);
		if (y_new == NULL)
			return (-1);

		y_old = y_dup(y, i);
		if (y_old == NULL) {
			FREE_OBJ(y_new);
			return (-1);
		}

		y->len = i;
		y->bitmask = 0x80 >> pt_bittbl[s[i] ^ *c];
		assert((*c & y->bitmask) != (s[i] & y->bitmask));
		bit = (*c & y->bitmask) != 0;
		assert(bit < 2);
		y->leaf[bit] = y_new;
		y->leaf[1-bit] = y_old;
		return (0);
	}
}

unsigned
PT_Lookup(const struct pt_y * const restrict root,
	  char * const restrict * const restrict strings,
	  const char * const restrict subject)
{
	size_t len;

	AN(strings);
	AN(subject);
	len = strlen(subject);

	for (const struct pt_y *y = root; y != NULL;) {
		unsigned char b;
		size_t l;

		CHECK_OBJ(y, PT_Y_MAGIC);
		l = y->off + y->len;
		if (l > len)
			return UINT_MAX;
		if (l == len && memcmp(subject, strings[y->idx], len) == 0)
			return y->idx;
		b = (y->bitmask & subject[l]) != 0;
		assert(b < 2);
		y = y->leaf[b];
	}

	return (UINT_MAX);
}

static int
pt_search(const struct pt_y * const restrict y,
	  char * const restrict * const restrict strings,
	  const unsigned char * restrict subject, size_t len,
	  struct match_data * const restrict match)
{
	size_t l;

	if (y == NULL)
		return (0);
	CHECK_OBJ(y, PT_Y_MAGIC);

	l = y->off + y->len;
	if (l > len)
		return (0);
	if (memcmp(subject + y->off, strings[y->idx] + y->off, y->len) != 0)
		return (0);

	if (strings[y->idx][l] == '\0') {
		if (match->n == match->limit)
			return (-1);
		match->indices[match->n] = y->idx;
		match->n++;
		if (y->idx < match->min)
			match->min = y->idx;
		if (y->idx > match->max)
			match->max = y->idx;
		if (l == len) {
			match->exact = y->idx;
			return (0);
		}
	}

	if (pt_search(y->leaf[0], strings, subject, len, match) != 0)
		return (-1);
	if (pt_search(y->leaf[1], strings, subject, len, match) != 0)
		return (-1);

	return (0);
}

int
PT_Prefixes(const struct pt_y * const restrict root,
	    char * const restrict * const restrict strings,
	    const char * const restrict subject,
	    struct match_data * const restrict match)
{
	size_t len;

	CHECK_OBJ_NOTNULL(match, MATCH_DATA_MAGIC);
	AN(match->indices);
	AN(match->limit);
	AN(strings);
	AN(subject);

	match->n = 0;
	match->min = UINT_MAX;
	match->max = 0;
	match->exact = UINT_MAX;

	len = strlen(subject);
	return (pt_search(root, strings, (unsigned char *)subject, len, match));
}

void
PT_Free(struct pt_y *y)
{
	if (y == NULL)
		return;
	CHECK_OBJ(y, PT_Y_MAGIC);
	PT_Free(y->leaf[0]);
	PT_Free(y->leaf[1]);
	FREE_OBJ(y);
}

static void
pt_print_tree(struct pt_y *y, struct vsb *sb, char **strings)
{
	CHECK_OBJ_NOTNULL(y, PT_Y_MAGIC);
	CHECK_OBJ_NOTNULL(sb, VSB_MAGIC);

	VSB_printf(sb, "node = %p\n", y);
	VSB_printf(sb, "leaf[0] = %p\n", y->leaf[0]);
	VSB_printf(sb, "leaf[1] = %p\n", y->leaf[1]);
	VSB_printf(sb, "idx = %u\n", y->idx);
	if (strings[y->idx] != NULL)
		VSB_printf(sb, "strings[idx] = %s\n", strings[y->idx]);
	VSB_printf(sb, "off = %u\n", y->off);
	VSB_printf(sb, "len = %u\n", y->len);
	VSB_printf(sb, "bitmask = 0x%02x\n\n", y->bitmask);

	if (y->leaf[0] != NULL)
		pt_print_tree(y->leaf[0], sb, strings);
	if (y->leaf[1] != NULL)
		pt_print_tree(y->leaf[1], sb, strings);
}

struct vsb *
PT_Dump(struct pt_y *root, char **strings)
{
	struct vsb *sb = VSB_new_auto();


	VSB_printf(sb, "root = %p\n\n", root);
	if (root != NULL) {
		AN(strings);
		pt_print_tree(root, sb, strings);
	}
	VSB_finish(sb);
	return (sb);
}

void
pt_stats(const struct pt_y * const restrict y,
	 char * const restrict * const restrict strings,
	 struct pt_stats * const restrict stats, unsigned depth)
{
	if (y == NULL)
		return;

	CHECK_OBJ(y, PT_Y_MAGIC);
	depth++;
	stats->nodes++;

	if (strings[y->idx][y->off + y->len] == '\0') {
		if (depth < stats->dmin)
			stats->dmin = depth;
		if (depth > stats->dmax)
			stats->dmax = depth;
		stats->davg += (depth - stats->davg) / (stats->terms + 1.);
		stats->terms++;
	}

	if (y->leaf[0] == NULL && y->leaf[1] == NULL) {
		stats->leaves++;
		return;
	}

	pt_stats(y->leaf[0], strings, stats, depth);
	pt_stats(y->leaf[1], strings, stats, depth);
}

void
PT_Stats(const struct pt_y * const restrict root,
	 char * const restrict * const restrict strings,
	 struct pt_stats * const restrict stats)
{
	CHECK_OBJ_NOTNULL(stats, PT_STATS_MAGIC);

	stats->nodes = 0;
	stats->leaves = 0;
	stats->terms = 0;
	stats->dmin = UINT64_MAX;
	stats->dmax = 0;
	stats->davg = 0.;
	stats->nodesz = sizeof(*root);

	pt_stats(root, strings, stats, 0);
}
